/*
 * avr_hd447800.h
 *
 * Created: 09.07.2021 16:35:11
 *  Author: mdudk
 */ 


#ifndef AVR_HD447800_H_
#define AVR_HD447800_H_

#include <util/delay.h> // do not forgot define F_CPU=xxxxxx in project/compiler/defined symbols
#include <avr/io.h>

/*
  driver for HD77480 / ST7066 or equivalents
  V2.1
 
  Library needs external _delay_us() function (like in "delay.h" library). Delays can be unprecise.
  Macros LCD_LINES a LCD_COLUMNS selects display variant.
  Library does not support push-pull or open drain communication yet (LCD contains at RS,RW,D0..D7 pins internal pullup current sources about 50-100uA, E pin must be pulled up externaly) 
  if you need save power call lcd_bus_sleep()/lcd_bus_wakeup() to deal with LCD driver internal "pullup" currents (which can significantly rise consumption)
  if using LCD without "hardwired" backlight, consider power LCD from GPIO pin (and when display unpowered use lcd_deinit_hw to setup all control pins to inputs)
  In that case dont forgot init display after powering up.
  
   
  for porting to another platform modify: 
	- lcd_init_hw() 
	- lcd_bus_outputs()
	- lcd_bus_inputs()
	- lcd_deinit_hw()
	- lcd_bus_set()
	- lcd_bus_read()
	- LCD_XX_L and LCD_XX_H macros
	- do not forgot provide _delay_us() function (not need to be precise)

   Typical (!) LCD pinout
    1 - GND
    2 - VCC
    3 - Contrast
    4 - RS
    5 - RW
    6 - E
    7 - D0 (not used)
    8 - D1 (not used)
    9 - D2 (not used)
   10 - D3 (not used)
   11 - D4
   12 - D5
   13 - D6
   14 - D7
   15 - Background (A) or neg.volt.generator output
   16 - Background (K)
 */

// ****************************** PART TO BE EDITED **********************************************
// TO BE IMPLEMENTED - select one variant of communication push-pull for 3.3V LCD, Open drain for 5V LCD with external pullup at enable pin)
// #define PUSHPULL_VARIANT
// #define OPENDRAIN_VARIANT

// uncomment if powering lcd from GPIO pin (feature for low power applications) - get acces to lcd_power_off()/on()/init() functions - can be simple handled manually 
// #define USE_POWER_PIN

// lcd_gotoxy() should work for 20x4, 16x2, 20x2 ... 16x1 is atypic and must be added, 8x2 not tested
// good informations here: http://web.alfredstate.edu/faculty/weimandn/lcd/lcd_addressing/lcd_addressing_index.html
#define LCD_LINES 2 // 1 , 2 or 4
#define LCD_COLUMNS 16 // 8 , 12 , 16 , 20 , 40 (not all are properly implemented)
// select used font (most common is 5x8px font)
//#define USED_FONT	LCD_FONT5X10
#define USED_FONT	LCD_FONT5X8

// define your own pinout
#define LCD_RS_PORT     PORTB
#define LCD_RW_PORT     PORTB
#define LCD_E_PORT      PORTB
#define LCD_D4_PORT     PORTB
#define LCD_D5_PORT     PORTB
#define LCD_D6_PORT     PORTB
#define LCD_D7_PORT     PORTB

#define LCD_RS_DDR		DDRB
#define LCD_RW_DDR		DDRB
#define LCD_E_DDR		DDRB
#define LCD_D4_DDR		DDRB
#define LCD_D5_DDR		DDRB
#define LCD_D6_DDR		DDRB
#define LCD_D7_DDR		DDRB

#define LCD_D4_INPUT	PINB
#define LCD_D5_INPUT	PINB
#define LCD_D6_INPUT	PINB
#define LCD_D7_INPUT	PINB

#define LCD_RS_PIN_NUMBER  0
#define LCD_RW_PIN_NUMBER  1
#define LCD_E_PIN_NUMBER   2
#define LCD_D4_PIN_NUMBER  3
#define LCD_D5_PIN_NUMBER  4
#define LCD_D6_PIN_NUMBER  5
#define LCD_D7_PIN_NUMBER  6

// define only if powering LCD from GPIO (feature for low power applications)
#ifdef USE_POWER_PIN
	#define LCD_PWR_PORT    PORTA
	#define LCD_PWR_PIN  	PIN0_bm
#endif

// *********************** PART NOT TO EDIT (unles you port or modify library) ************************************
// If need, you can change delays and timeouts. All in us.
#define LCD_STARTUP_DELAY1 15000
#define LCD_STARTUP_DELAY2 4100
#define LCD_STARTUP_DELAY3 100
#define LCD_E_DELAY 1
#define LCD_COMMAND_DELAY 40
#define LCD_BUSY_TIMEOUT 500 // maximum attempts to read busy flags

// macros for easy compatibility AVR with AVR1 series and other MCUs - defining pin masks...
#define LCD_RS_PIN  (1<<LCD_RS_PIN_NUMBER)
#define LCD_RW_PIN  (1<<LCD_RW_PIN_NUMBER)
#define LCD_E_PIN   (1<<LCD_E_PIN_NUMBER)
#define LCD_D4_PIN  (1<<LCD_D4_PIN_NUMBER)
#define LCD_D5_PIN  (1<<LCD_D5_PIN_NUMBER)
#define LCD_D6_PIN  (1<<LCD_D6_PIN_NUMBER)
#define LCD_D7_PIN  (1<<LCD_D7_PIN_NUMBER)

// macros for "push pull" variant
#define LCD_RS_H    LCD_RS_PORT |=	LCD_RS_PIN
#define LCD_RS_L    LCD_RS_PORT &=  ~LCD_RS_PIN
#define LCD_RW_H    LCD_RW_PORT |=	LCD_RW_PIN
#define LCD_RW_L    LCD_RW_PORT &=  ~LCD_RW_PIN
#define LCD_E_H     LCD_E_PORT |=	LCD_E_PIN
#define LCD_E_L     LCD_E_PORT &=  ~LCD_E_PIN
#define LCD_D4_H    LCD_D4_PORT |=	LCD_D4_PIN
#define LCD_D4_L    LCD_D4_PORT &=  ~LCD_D4_PIN
#define LCD_D5_H    LCD_D5_PORT |=	LCD_D5_PIN
#define LCD_D5_L    LCD_D5_PORT &=  ~LCD_D5_PIN
#define LCD_D6_H    LCD_D6_PORT |=	LCD_D6_PIN
#define LCD_D6_L    LCD_D6_PORT &=  ~LCD_D6_PIN
#define LCD_D7_H    LCD_D7_PORT |=	LCD_D7_PIN
#define LCD_D7_L    LCD_D7_PORT &=  ~LCD_D7_PIN

// input/output macros
#define LCD_RS_OUT  LCD_RS_DDR  |= LCD_RS_PIN
#define LCD_RS_IN	LCD_RS_DDR &=  ~LCD_RS_PIN
#define LCD_RW_OUT  LCD_RW_DDR |= LCD_RW_PIN
#define LCD_RW_IN	LCD_RW_DDR  &=  ~LCD_RW_PIN
#define LCD_E_OUT	LCD_E_DDR  |= LCD_E_PIN
#define LCD_E_IN	LCD_E_DDR &=  ~LCD_E_PIN

#define LCD_D4_OUT  LCD_D4_DDR |= LCD_D4_PIN
#define LCD_D4_IN   LCD_D4_DDR  &=  ~LCD_D4_PIN
#define LCD_D5_OUT  LCD_D5_DDR |= LCD_D5_PIN
#define LCD_D5_IN   LCD_D5_DDR  &=  ~LCD_D5_PIN
#define LCD_D6_OUT  LCD_D6_DDR |= LCD_D6_PIN
#define LCD_D6_IN   LCD_D6_DDR  &=  ~LCD_D6_PIN
#define LCD_D7_OUT  LCD_D7_DDR |= LCD_D7_PIN
#define LCD_D7_IN   LCD_D7_DDR  &=  ~LCD_D7_PIN


// macros for "open drain" variant - NOT IMPLEMENTED YET

// pin read macros
#define LCD_D4_VALUE	(LCD_D4_INPUT & LCD_D4_PIN)
#define LCD_D5_VALUE	(LCD_D5_INPUT & LCD_D5_PIN)
#define LCD_D6_VALUE	(LCD_D6_INPUT & LCD_D6_PIN)
#define LCD_D7_VALUE	(LCD_D7_INPUT & LCD_D7_PIN)

// commands
#define LCD_DISPLAY_CLEAR           (0b1<<0)

#define LCD_RETURN_HOME             (0b1<<1)

#define LCD_ENTRY_MODE_SET          (0b1<<2)
	#define LCD_DECREMENT           (0b0<<1)
	#define LCD_INCREMENT           (0b1<<1)
	#define LCD_ENTRY_W_SHIFT      (0b1<<0)
	#define LCD_DISPLAY_NOSHIFT    (0b0<<0)

#define LCD_DISPLAY_ONOFF   (0b1<<3)
	#define LCD_ON             (0b1<<2)
	#define LCD_OFF            (0b0<<2)
	#define LCD_CURSOR_ON      (0b1<<1)
	#define LCD_CURSOR_OFF     (0b0<<1)
	#define LCD_BLINK_ON       (0b1<<0)
	#define LCD_BLINK_OFF      (0b0<<0)

#define LCD_CURSOR_OR_DISPLAY_SHIFT		(0b1<<4)
	#define LCD_DISPLAY_SHIFT          (0b1<<3)
	#define LCD_CURSOR_SHIFT           (0b0<<3)
	#define LCD_SHIFT_RIGHT            (0b1<<2)
	#define LCD_SHIFT_LEFT             (0b0<<2)

#define LCD_FUNCTION_SET    (0b1<<5)
 #define LCD_8BITS          (0b1<<4)
 #define LCD_4BITS          (0b0<<4)
 #define LCD_2LINES         (0b1<<3)
 #define LCD_1LINES         (0b0<<3)
 #define LCD_FONT5X10       (0b1<<2)
 #define LCD_FONT5X8        (0b0<<2)

#define LCD_SET_CGRAM       (0b1<<6)

#define LCD_SET_DDRAM       (0b1<<7)


void lcd_bus_set(uint8_t data); 		// write output values on D7..D4 bus
void lcd_init_hw(void);					// init GPIOs etc.
void lcd_deinit_hw(void);				// set all LCD control GPIOs as floating input (use if display is powered off or you risk powering it from bus)
void lcd_bus_outputs(void);				// set D7..D4 as outputs (for writing to LCD)
void lcd_bus_inputs(void);				// set D7..D4 as inputs (for reading from LCD)
uint8_t lcd_bus_read(void);				// read 4bit value from D7..D4
void lcd_e_tick(void);					// generate positive pulse/clock at Enable pin
void lcd_command(uint8_t command);		// sends command to LCD and waits until busy
void lcd_data(uint8_t data);			// sends data to LCD and waits until busy
uint8_t lcd_read(void);					// read 8bit value from LCD (MSB is busy flag, LSBs address counter)
uint8_t lcd_busy_wait(void);			// wait until LCD is busy or timeout
void lcd_bus_sleep(void);				// set LCD pins to High level to cancel internal pullup currents (for low power applications)
void lcd_bus_wakeup(void);				// set LCD pins back to default state (call prior to lcd operations if previously called lcd_bus_sleep)

void lcd_init(void);									// initialisate HW and display
void lcd_gotoxy(uint8_t column, uint8_t line);			// set cursor at desired position
void lcd_puts(char* text);								// prints string to LCD
void lcd_store_symbol(uint8_t pos, uint8_t* charmap); 	// write users own charaters into LCD CGRAM
//void _delay_us(uint16_t Delay);

#define lcd_putchar(X)		lcd_data(X)
#define lcd_busy() 			(lcd_read() & 0b1<<7)  			// check busy status flag
#define lcd_clear() 		lcd_command(LCD_DISPLAY_CLEAR) 	// clears entire display
#define lcd_home() 			lcd_command(LCD_RETURN_HOME)	// set cursor and display shift to home position
#define lcd_cursor_on() 	lcd_command(LCD_DISPLAY_ONOFF | LCD_ON | LCD_CURSOR_ON | LCD_BLINK_OFF)		// activate underscore cursor
#define lcd_cursor_blink() 	lcd_command(LCD_DISPLAY_ONOFF | LCD_ON | LCD_CURSOR_OFF | LCD_BLINK_ON) 	// activate blinking block 5x8 cursor
#define lcd_cursor_off() 	lcd_command(LCD_DISPLAY_ONOFF | LCD_ON | LCD_CURSOR_OFF | LCD_BLINK_OFF) 	// deactivate both cursor (underscore and blink)

#ifdef USE_POWER_PIN
	#define lcd_pwr_pin_init(void) LCD_PWR_PORT.DIRSET = LCD_PWR_PIN
	#define lcd_power_off(void) LCD_PWR_PORT.OUTCLR = LCD_PWR_PIN
	#define lcd_power_oon(void) LCD_PWR_PORT.OUTSET = LCD_PWR_PIN
#endif

#define lcd_delay_us(D) _delay_us(D)


#endif /* AVR_HD447800_H_ */